//
//
// HMC Lion password board prototype
//
// set lfuse to 0x5E for 20 MHz xtal
//
// Lincoln Craven-Brightman
// 19/11/18
//

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define input_pins PINA
#define input_port PORTA
#define digit_pin (1 << PA0)
#define go_pin (1 << PA1)
#define knocks_pin (1 << PA2)
#define pass_reset_pin (1 << PA7)

#define led_pin (1 << PA3)
#define led_port PORTA
#define led_direction DDRA

#define pass_out_pin (1 << PB2)
#define pass_out_port PORTB
#define pass_out_direction DDRB

#define serial_pin_out (1 << PA5)
#define serial_port PORTA
#define serial_direction DDRA

#define digits 3

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
}

void get_char(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
   //
   // read character into rxbyte on pins pin
   //    assumes line driver (inverts bits)
   //
   *rxbyte = 0;
   while (pin_test(*pins,pin)) {
      //
      // wait for start bit
      //
   }
   //
   // delay to middle of first data bit
   //
   half_bit_delay();
   bit_delay();
   //
   // unrolled loop to read data bits
   //
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 0);
   else
      *rxbyte |= (0 << 0);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 1);
   else
      *rxbyte |= (0 << 1);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 2);
   else
      *rxbyte |= (0 << 2);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 3);
   else
      *rxbyte |= (0 << 3);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 4);
   else
      *rxbyte |= (0 << 4);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 5);
   else
      *rxbyte |= (0 << 5);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 6);
   else
      *rxbyte |= (0 << 6);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 7);
   else
      *rxbyte |= (0 << 7);
   //
   // wait for stop bit
   //
   bit_delay();
   half_bit_delay();
}

int main(void) {
   //
   // main
   //
   static char password[] = {8, 6, 3};
   static char knocks;
   static char buffer[digits];
   static int digit;
   static int dig_on;
   static int lp_i;
   static int match;
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize pins
   //
   set(pass_out_port, pass_out_pin);
   set(input_port, digit_pin); // pull digit signal high
   set(input_port, go_pin); // pull go signal high
   set(input_port, knocks_pin); // pull knocks signal high
   set(input_port, pass_reset_pin); // pull password reset signal high
   output(led_direction, led_pin);
   output(pass_out_direction, pass_out_pin);

   set(serial_port, serial_pin_out); 
   output(serial_direction, serial_pin_out);

   //
   // initialize buffer
   //
   digit = 0;
   dig_on = 0;
   for (lp_i = 0; lp_i < digits; lp_i ++) {
      buffer[lp_i] = 0;
   }
   //
   // boot delay
   //
   set(led_port, led_pin);
   _delay_ms(500);
   clear(led_port, led_pin);
   //
   // main loop
   //
   while (1) {
      // if (!pin_test(input_pins, pass_reset_pin)) {
      //    // reset digit, buffer
      //    digit = 0;
      //    dig_on = 0;
      //    for (lp_i = 0; lp_i < digits; lp_i ++) {
      //       buffer[lp_i] = 0;
      //    }
      // }
      if (!pin_test(input_pins, go_pin)) {
         _delay_ms(75);
         set(led_port, led_pin);
         while (!pin_test(input_pins, go_pin)) {
            // Wait until not counting knocks
         }
         
         get_char(&input_pins, knocks_pin, &knocks);
         if (knocks > 0 && knocks < 10) {
            buffer[digit] = knocks;
         }
         clear(led_port, led_pin);
         digit = 0;

         // Check for match
         match = 1;
         for (lp_i = 0; lp_i < digits; lp_i ++) {
            put_char(&serial_port, serial_pin_out, buffer[lp_i]);
            char_delay();
            if (buffer[lp_i] != password[lp_i]) {
               match = 0;
            }
         }
         if (match) {
            clear(pass_out_port, pass_out_pin);
            set(led_port, led_pin);
            _delay_ms(500);
            clear(led_port, led_pin);
            _delay_ms(500);
            set(led_port, led_pin);
            _delay_ms(500);
            clear(led_port, led_pin);
            _delay_ms(500);
            set(led_port, led_pin);
            _delay_ms(500);
         }
         else {
            set(pass_out_port, pass_out_pin);
            clear(led_port, led_pin);
         }
      }
      else if (!pin_test(input_pins, digit_pin)) {
         // increment digits on rising edge
         if (digit < (digits - 1) && !dig_on) {
            digit ++;
            _delay_ms(50);
            dig_on = 1;
            //
            // send digit number to serial probe
            //
            put_char(&serial_port, serial_pin_out, 'd');
            char_delay();
            put_char(&serial_port, serial_pin_out, (char) digit);
            char_delay();
         }
      }
      else {
         dig_on = 0;
      } 
   }
}
